#!/usr/bin/env python

"""Identify the "best" observation model by treating the accuracy of the peak
timing forecasts, when plotted against the time to the true peak, as an ROC
and calculating the AUC ("area under curve") for each calendar year.

This is not a technically valid assumption, since the curves are not monotonic
and the abscissa is not equivalent to the unit interval, but it serves to
provide a pragmatic metric against which observation models can be compared.
"""

from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

import argparse
import datetime
import glob
import h5py
import numpy as np
import os.path
import sys


def date_of_string(string):
    """Convert a string into a ``datetime.date`` instance."""
    return datetime.datetime.strptime(string, "%Y-%m-%d %H:%M:%S").date()


def auc_scores(f, args):
    """Calculate the AUC scores from a single simulation, for all reported
    accuracy tolerances (i.e., +/- XX days).
    """
    # Determine the timing of the true peak.
    obs = f['/data/obs'][()]
    max_row = np.where(obs['value'] == max(obs['value']))
    true_peak = date_of_string(obs['date'][max_row][0])

    # Calculate the ROC and AUC for each tolerance, over the 8 week
    # interval prior to the true peak.
    tacc = f['/data/peak_time_acc'][()]
    fs_dates = np.array([date_of_string(s) for s in tacc['fs_date']])
    mask = np.array([0 <= (true_peak - d).days <= args.weeks * 7
                     for d in fs_dates])
    tacc = tacc[mask]
    tolns = set(tacc['toln'])

    return {t: np.sum(tacc['acc'][tacc['toln'] == t]) for t in tolns}


def key_info(f, args):
    obs = f['/data/obs'][()]
    year = date_of_string(obs['date'][0]).year
    src = obs['source'][0]

    # Determine the key that identifies this observation model.
    units = f['/meta/param/obs/'].keys()
    if len(units) != 1:
        raise ValueError("Too many observation units: {}".format(units))

    om = f['/meta/param/obs/{}'.format(units[0])]
    param_names = om.keys()
    pr_key = filter(lambda n: n.startswith("pr_"), param_names)[0]
    bg_key = filter(lambda n: "_bg" in n, param_names)[0]
    obs_pr = "{}".format(om[pr_key][()])
    obs_bg = "{}".format(om[bg_key][()])
    disp = "{}".format(om['disp'][()])

    if src == 'VDH':
        # Account for changes in VDH background rate from year to year.
        vdh_bg = {2009: 14, 2010: 15, 2011: 32, 2012: 28, 2013: 32,
                  2014: 46, 2015: 58}
        obs_bg = int(obs_bg) - vdh_bg[year]

    return {'year': year, 'pr': obs_pr, 'bg': obs_bg, 'disp': disp,
            'src': src}


def print_auc_table(pattern, args, outfile):
    header = "src pr bg disp year toln auc"
    print(header, file=outfile)

    fmt = "{src} {pr} {bg} {disp} {year} {toln} {auc}"
    for filename in sorted(glob.iglob(pattern)):
        with h5py.File(filename, 'r') as f:
            aucs = auc_scores(f, args)
            info = key_info(f, args)
            for toln in aucs:
                line = fmt.format(toln=toln, auc=aucs[toln], **info)
                print(line, file=outfile)


def get_parser():
    """Return the command-line argument parser for this script."""
    p = argparse.ArgumentParser()
    p.add_argument(
        '--weeks', '-w', type=int, default=8,
        help='The duration of the AUC period (default: %(default)s)')
    p.add_argument(
        'data_dir', type=str,
        help='The directory containing the forecasting outputs.')

    return p


def main(args=None):
    # Parse the command-line arguments.
    parser = get_parser()
    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    dir_name = os.path.basename(os.path.abspath(args.data_dir))
    output_file = os.path.join(args.data_dir, 'auc-{}.ssv'.format(dir_name))
    pattern = "{}/*.hdf5".format(args.data_dir)
    with open(output_file, 'w') as f:
        print_auc_table(pattern, args, f)

    return 0


if __name__ == "__main__":
    sys.exit(main())
