"""Observation models and data sources for metropolitan Melbourne."""

from __future__ import absolute_import, division, print_function
# Note: unicode_literals produces NumPy errors, since dtype field names cannot
#       be unicode (https://github.com/numpy/numpy/issues/2407).
# from __future__ import unicode_literals

import datetime
import numpy as np
import scipy.stats


def _date_col(text, fmt='%Y-%m-%d'):
    """Convert date strings to datetime.datetime instances. This is a
    convenience function intended for use with, e.g., numpy.loadtxt().
    """
    return datetime.datetime.strptime(text.decode('utf-8'), fmt)


class GPSS(object):
    """General Practitioner Sentinel Surveillance (now VicSPIN)."""

    obs_unit = "ILI Cases"
    obs_period = 7

    # Ignore the final week of 2011, with 174 patients and a *huge* 2.9% ILI.
    min_patients = 200

    @classmethod
    def populate(cls, params, args):
        """Add observation model parameters to the simulation parameters."""
        params['obs'][cls.obs_unit] = {
            'expect_fn': cls.expect,
            'log_llhd_fn': cls.log_llhd,
            'pr_pres_inf': args['gpss_pr'],
            'obs_bg': args['gpss_bg'],
            'disp': args['gpss_disp']
        }

    @classmethod
    def obs(cls, filename, year=None):
        col_types = [('year', np.int32), ('week', np.int8), ('from', '|O4'),
                     ('to', '|O4'), ('patients', np.int32), ('ili', np.int32),
                     ('pcnt', np.float)]
        col_convs = {2: _date_col, 3: _date_col}
        df = np.loadtxt(filename, skiprows=1, dtype=col_types,
                        converters=col_convs)
        if year is not None:
            df = df[df['year'] == year]

        nrows = df.shape[0]
        return [{'date': df['to'][i],
                 'value': df['ili'][i],
                 'denominator': df['patients'][i],
                 'unit': cls.obs_unit,
                 'period': cls.obs_period,
                 'source': "GPSS"}
                for i in range(nrows)
                if df['patients'][i] >= cls.min_patients]

    @staticmethod
    def expect(params, op, period, pr_inf):
        """Determine the expected value for a given infection probability.

        :param params: The observation model parameters.
        :param op: The observation model parameters.
        :param period: The duration of the observation period (in days).
        :param pr_inf: The probability of an individual becoming infected
            during the observation period.
        """
        n = params['epifx']['popn_size']
        return op['obs_bg'] + pr_inf * op['pr_pres_inf'] * n

    @staticmethod
    def log_llhd(params, op, obs, pr_indiv, curr, hist):
        """Calculate the log-likelihood of obtaining one or more observations
        from each particle.

        :param params: The simulation parameters.
        :param op: The observation model parameters.
        :param obs: The list of observations for the current time-step.
        :param pr_indiv: A 2 x N matrix (for N particles) of the probability
            of an individual being uninfected and infected, respectively.
        :param curr: The particle state vectors.
        :param hist: The particle state histories, indexed by observation
            period.
        """
        obs_count = obs['value']
        # The (mean) background rate.
        obs_bg = op['obs_bg']
        # Observations are obtained from the entire population.
        size = params['epifx']['popn_size']
        # ILI presentation probabilities (uninfected and infected).
        pr_pres_inf = op['pr_pres_inf']
        pr_ili = np.array([obs_bg / float(size), pr_pres_inf])
        # Calculate the net probability of ILI presentations.
        pr_obs = np.sum(pr_ili[:, None] * pr_indiv, axis=0)
        # Calculate the expected number of failures (i.e., cases).
        mu = pr_obs * size
        # Extract the over-dispersion parameter.
        nb_k = op['disp']
        nb_pr = nb_k / (nb_k + mu)
        nbinom = scipy.stats.nbinom(nb_k, nb_pr)
        return nbinom.logpmf(obs_count)


class MMDS(object):
    """Melbourne Medical Deputising Service (now National Home Doctor
    Service).
    """

    obs_unit = "ILI %"
    obs_period = 7

    # Ignore two outliers: the first exceeds the *actual* 2013 epidemic peak,
    # while the second *almost* exceeds the 2014 peak in *January*.
    ign_dates = [datetime.datetime(2013, 11, 24),
                 datetime.datetime(2014, 1, 12)]

    @classmethod
    def populate(cls, params, args):
        """Add observation model parameters to the simulation parameters."""
        params['obs'][cls.obs_unit] = {
            'expect_fn': cls.expect,
            'log_llhd_fn': cls.log_llhd,
            'pr_call_inf': args['mmds_pr'],
            'obs_bg_frac': args['mmds_bg'],
            'disp': args['mmds_disp']
        }

    @classmethod
    def obs(cls, filename, year=None):
        col_types = [('year', np.int32), ('week', np.int8), ('from', '|O4'),
                     ('to', '|O4'), ('calls', np.int32), ('ili', np.int32),
                     ('pcnt', np.float64)]
        col_convs = {2: _date_col, 3: _date_col}
        df = np.loadtxt(filename, skiprows=1, dtype=col_types,
                        converters=col_convs)
        if year is not None:
            df = df[df['year'] == year]

        nrows = df.shape[0]
        return [{'date': df['to'][i],
                 'value': df['pcnt'][i],
                 'denominator': df['calls'][i],
                 'unit': cls.obs_unit,
                 'period': cls.obs_period,
                 'source': "MMDS"}
                for i in range(nrows)
                if df['to'][i] not in cls.ign_dates]

    @staticmethod
    def expect(params, op, period, pr_inf):
        """Determine the expected value for a given infection probability.

        :param params: The observation model parameters.
        :param op: The observation model parameters.
        :param period: The duration of the observation period (in days).
        :param pr_inf: The probability of an individual becoming infected
            during the observation period.
        """
        n = params['epifx']['popn_size']
        return 100 * (op['obs_bg_frac'] + pr_inf * op['pr_call_inf'] * n)

    @staticmethod
    def log_llhd(params, op, obs, pr_indiv, curr, hist):
        """Calculate the log-likelihood of obtaining one or more observations
        from each particle.

        :param params: The simulation parameters.
        :param op: The observation model parameters.
        :param obs: The list of observations for the current time-step.
        :param pr_indiv: A 2 x N matrix (for N particles) of the probability
            of an individual being uninfected and infected, respectively.
        :param curr: The particle state vectors.
        :param hist: The particle state histories, indexed by observation
            period.
        """
        # Extract the number of ILI calls and total calls.
        ili_frac = obs['value'] / 100.0
        total_calls = obs['denominator']
        obs_count = np.round(ili_frac * total_calls)
        obs_bg = np.round(op['obs_bg_frac'] * total_calls)
        # Observations are obtained from the telephone calls.
        size = total_calls
        # ILI consultation probabilities (uninfected and infected).
        pr_call_inf = op['pr_call_inf'] * params['epifx']['popn_size']
        pr_ili = np.array([obs_bg / float(size), pr_call_inf])
        # Calculate the net probability of ILI consultations.
        pr_obs = np.sum(pr_ili[:, None] * pr_indiv, axis=0)
        # Calculate the expected number of failures (i.e., cases).
        mu = pr_obs * size
        # Extract the over-dispersion parameter.
        nb_k = op['disp']
        nb_pr = nb_k / (nb_k + mu)
        nbinom = scipy.stats.nbinom(nb_k, nb_pr)
        return nbinom.logpmf(obs_count)


class VDH(object):
    """Victorian Department of Health influenza notifications."""

    obs_unit = "Notifications"
    obs_period = 7

    @classmethod
    def populate(cls, params, args):
        """Add observation model parameters to the simulation parameters."""
        params['obs'][cls.obs_unit] = {
            'expect_fn': cls.expect,
            'log_llhd_fn': cls.log_llhd,
            'pr_notif_inf': args['vdh_pr'],
            'obs_bg': args['vdh_bg'],
            'disp': args['vdh_disp']
        }

    @classmethod
    def obs(cls, filename, year=None):
        # The notifications data file may include the week number column,
        # so assume that it is present and try to parse the data file.
        col_types = [('year', np.int32), ('week', np.int8), ('from', '|O4'),
                     ('to', '|O4'), ('count', np.int32)]
        col_convs = {2: _date_col, 3: _date_col}
        try:
            df = np.loadtxt(filename, skiprows=1, dtype=col_types,
                            converters=col_convs)
        except ValueError:
            # If an invalid value is encountered, assume that the week number
            # column is not present and try to parse the data file again.
            col_types = [('year', np.int32), ('from', '|O4'), ('to', '|O4'),
                         ('count', np.int32)]
            col_convs = {1: _date_col, 2: _date_col}
            df = np.loadtxt(filename, skiprows=1, dtype=col_types,
                            converters=col_convs)

        if year is not None:
            df = df[df['year'] == year]

        nrows = df.shape[0]
        return [{'date': df['to'][i],
                 'value': df['count'][i],
                 'unit': cls.obs_unit,
                 'period': cls.obs_period,
                 'source': "VDH"}
                for i in range(nrows)]

    @staticmethod
    def expect(params, op, period, pr_inf):
        """Determine the expected value for a given infection probability.

        :param params: The observation model parameters.
        :param op: The observation model parameters.
        :param period: The duration of the observation period (in days).
        :param pr_inf: The probability of an individual becoming infected
            during the observation period.
        """
        n = params['epifx']['popn_size']
        return op['obs_bg'] + pr_inf * op['pr_notif_inf'] * n

    @staticmethod
    def log_llhd(params, op, obs, pr_indiv, curr, hist):
        """Calculate the log-likelihood of obtaining one or more observations
        from each particle.

        :param params: The simulation parameters.
        :param op: The observation model parameters.
        :param obs: The list of observations for the current time-step.
        :param pr_indiv: A 2 x N matrix (for N particles) of the probability
            of an individual being uninfected and infected, respectively.
        :param curr: The particle state vectors.
        :param hist: The particle state histories, indexed by observation
            period.
        """
        obs_count = obs['value']
        # The (mean) background rate.
        obs_bg = op['obs_bg']
        # Observations are obtained from the entire population.
        size = params['epifx']['popn_size']
        # Notification probabilities (uninfected and infected).
        pr_notif_inf = op['pr_notif_inf']
        pr_ili = np.array([obs_bg / float(size), pr_notif_inf])
        # Calculate the net probability of flu notifications.
        pr_obs = np.sum(pr_ili[:, None] * pr_indiv, axis=0)
        # Calculate the expected number of failures (i.e., cases).
        mu = pr_obs * size
        # Extract the over-dispersion parameter.
        nb_k = op['disp']
        nb_pr = nb_k / (nb_k + mu)
        nbinom = scipy.stats.nbinom(nb_k, nb_pr)
        return nbinom.logpmf(obs_count)
