#!/usr/bin/env python
#
# Generate forecasts from multiple data sources for metropolitan Melbourne.
#

from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

import argparse
import datetime
import gc
import logging
import numpy as np
import resource
import sys

import pypfilt
import epifx

import melbn


def params_for(args, p_exp, px_scale, **kwargs):
    """Return a default set of parameters for the single-strain SEIR model.
    Additional keyword arguments are assumed to define other parameters.

    :param args: The dictionary of parsed command-line arguments.
    :param p_exp: The daily probability of seeding an exposure event in a
        naive population.
    :param px_scale: The ratio of particles to seeded exposures.
    :param kwargs: Values for individual simulation parameters.
    """
    params = epifx.default_params(p_exp, px_scale)
    params['epifx']['one_prng'] = True

    # Define the observation model parameters.
    if args['gpss']:
        melbn.GPSS.populate(params, args)
    if args['mmds']:
        melbn.MMDS.populate(params, args)
    if args['vdh']:
        melbn.VDH.populate(params, args)

    # Set the scale of the stochastic noise.
    popn_size = params['epifx']['popn_size']
    params['prior']['noise_flow'] = lambda r: 0.025 * np.sqrt(1.0 / popn_size)
    params['prior']['noise_param'] = lambda r: 5e-3

    # Set the output directory, if specified.
    if args['out_dir'] is not None:
        params['out_dir'] = args['out_dir']

    # Apply any parameters that were passed as keyword arguments.
    for k in kwargs:
        if k in params:
            params[k] = kwargs[k]
        else:
            print("Warning: ignoring unknown parameter {}".format(k))

    return params


def output_file(args, params):
    """Determine the name of the simulation output file.

    :param args: The parsed command-line arguments.
    :param params: The simulation parameters.
    """
    srcs, prs, bgs, disps = [], [], [], []
    if args['gpss']:
        srcs.append("gpss")
        prs.append("{}".format(args['gpss_pr']))
        bgs.append("{}".format(args['gpss_bg']))
        disps.append("{}".format(args['gpss_disp']))
    if args['mmds']:
        srcs.append("mmds")
        prs.append("{}".format(args['mmds_pr']))
        bgs.append("{}".format(args['mmds_bg']))
        disps.append("{}".format(args['mmds_disp']))
    if args['vdh']:
        srcs.append("vdh")
        prs.append("{}".format(args['vdh_pr']))
        bgs.append("{}".format(args['vdh_bg']))
        disps.append("{}".format(args['vdh_disp']))
    src = "+".join(srcs)
    pr = "+".join(prs)
    bg = "+".join(bgs)
    disp = "+".join(disps)

    if args['suffix'] is not None:
        return "{}-{}-{}-pr-{}-bg-{}-disp-{}-{}.hdf5".format(
            src, args['year'], params['model'].__name__, pr, bg, disp,
            args['suffix'])
    else:
        return "{}-{}-{}-pr-{}-bg-{}-disp-{}.hdf5".format(
            src, args['year'], params['model'].__name__, pr, bg, disp)


def run(args, num_fs=None):
    """Compare forecasts obtained from the single-strain SEIR model against
    Victoria surveillance data, over a range of values for several key
    forecasting parameters.

    :param args: The dictionary of parsed command-line arguments.
    :param num_fs: Forecast from the last N observations, if positive;
        forecast from all observations, if negative; forecast from mid March
        until the epidemic peak, if ``None``.
    """
    logger = logging.getLogger(__name__)

    gc.collect()
    logger.debug("Memory usage: {} (kb)".format(
        resource.getrusage(resource.RUSAGE_SELF).ru_maxrss))

    if args['quick']:
        px_scale = 2
        p_exp = 1.0 / 50
    else:
        px_scale = 100
        p_exp = 1.0 / 36

    params = params_for(args, p_exp, px_scale, prng_seed=3001)
    # Keep eta fixed at 1 (i.e., enforce homogeneous mixing).
    params['prior']['eta'] = lambda r, size=None: 1.0
    params['epifx']['param_min'][3] = 1.0
    params['epifx']['param_max'][3] = 1.0

    start = datetime.datetime(args['year'], 1, 1)
    until = datetime.datetime(args['year'] + 1, 1, 1)
    data_file = output_file(args, params)

    # Load the observations for each of the chosen data sources.
    obs = []
    if args['gpss']:
        obs.append(melbn.GPSS.obs(args['gpss_file'], year=start.year))
    if args['mmds']:
        obs.append(melbn.MMDS.obs(args['mmds_file'], year=start.year))
    if args['vdh']:
        obs.append(melbn.VDH.obs(filename=args['vdh_file'], year=start.year))
    all_obs = sum(obs, [])
    obs_dates = sorted(set(o['date'] for o in all_obs))

    sim_start = datetime.datetime.now()
    logger.debug("forecast() beginning at {}".format(
        sim_start.strftime("%H:%M:%S")))

    # Pick all observation dates from mid March until the epidemic peak.
    # First, identify the time and size of the peak for each data stream.
    peak_sizes = [max(o['value'] for o in stream) for stream in obs]
    peak_dates = []
    for ix, size in enumerate(peak_sizes):
        peak_obs = [o for o in obs[ix] if o['value'] == size][0]
        peak_dates.append(peak_obs['date'])

    # Determine how often observations are reported (in days).
    if (obs_dates[1] - obs_dates[0]).days == 1:
        # Generate weekly forecasts when given daily observations, because
        # daily forecasts provide no additional insight and consume a huge
        # amount of computation time and disk space.
        obs_dates = [o for o in obs_dates if o.strftime('%w') == '0']

    # By default, start forecasting from the third observation in March.
    # But if there are no observations in March (e.g., GPSS) then start at
    # the earliest observation date.
    march_dates = [d for d in obs_dates if d.month == 3]
    if len(march_dates) >= 3:
        dmin = [d for d in obs_dates if d.month == 3][2]
    elif len(march_dates) > 0:
        dmin = [d for d in obs_dates if d.month == 3][0]
    else:
        dmin = min(obs_dates)

    # Ensure the final forecast is no earlier than the peak, which may or
    # may not coincide with a weekly forecasting date.
    dmax = max(peak_dates) + datetime.timedelta(days=6)
    fs_dates = [d for d in obs_dates if dmin <= d <= dmax]

    summary = epifx.Summary(params, all_obs)
    pypfilt.forecast(params, start, until, obs, fs_dates, summary,
                     filename=data_file)
    logger.debug("forecast() finishing at {}".format(
        datetime.datetime.now().strftime("%H:%M:%S")))

    gc.collect()
    logger.debug("Memory usage: {} (kb)".format(
        resource.getrusage(resource.RUSAGE_SELF).ru_maxrss))


def get_parser():
    """Return the command-line argument parser for this script."""

    # The main command-line parser.
    p = argparse.ArgumentParser()

    log = p.add_mutually_exclusive_group()
    log.add_argument(
        '-d', '--debug',
        help='Enable debugging output',
        action='store_const', dest='loglevel',
        const=logging.DEBUG, default=logging.INFO)
    log.add_argument(
        '-q', '--quiet',
        help='Suppress logging output',
        action='store_const', dest='loglevel',
        const=logging.WARNING)

    # The parser for forecasting options.
    fp = p.add_argument_group('Forecasting')
    fp.add_argument(
        '--quick', action='store_true',
        help='Perform a single run with minimal particles')
    fp.add_argument(
        '--parts', type=int, default=1,
        help='The number of simulation partitions')
    fp.add_argument(
        'choice', type=int, metavar='NUM', nargs='?', default=None,
        help='The partition to simulate')

    # The parser for input options.
    ip = p.add_argument_group(title='Input')
    ip.add_argument(
        '--gpss-file', type=str, metavar='FILE',
        default='gpss-weekly-counts.ssv',
        help='GPSS ILI consultations data file')
    ip.add_argument(
        '--mmds-file', type=str, metavar='FILE',
        default='mmds-weekly-counts.ssv',
        help='MMDS ILI consultations data file')
    ip.add_argument(
        '--vdh-file', type=str, metavar='FILE',
        default='vdh-weekly-counts.ssv',
        help='VDH notifications data file')
    ip.add_argument(
        '--no-gpss', action='store_false', dest='gpss',
        help='Do not generate GPSS forecasts')
    ip.add_argument(
        '--no-mmds', action='store_false', dest='mmds',
        help='Do not generate MMDS forecasts')
    ip.add_argument(
        '--no-vdh', action='store_false', dest='vdh',
        help='Do not generate VDH forecasts')

    # The parser for output options.
    op = p.add_argument_group(title='Output')
    op.add_argument(
        '--out-dir', type=str, default=None, metavar='DIR',
        help='The directory where forecast outputs will be saved')
    op.add_argument(
        '--suffix', type=str, default=None,
        help='The suffix for output file names')

    return p


def main(argv=None):
    # Parse the command-line arguments.
    parser = get_parser()
    if argv is None:
        args = vars(parser.parse_args())
    else:
        args = vars(parser.parse_args(argv))

    logging.basicConfig(level=args['loglevel'])

    # Generate the specified forecasts.
    partitions = args['parts']
    choice = args['choice']

    if partitions < 1:
        parser.error('argument --parts must be positive')
    if choice is None and partitions == 1:
        # There is only one choice, don't require the user to pick it.
        choice = 1
    elif not (choice >= 1 and choice <= partitions):
        msg = 'partition number must be between 1 and {}'.format(
            partitions)
        parser.error(msg)

    # Use zero-based counting for partitions.
    counter = 0
    choice -= 1

    # Simulation parameters common to all forecasts.
    years = [2009, 2010, 2011, 2012, 2013, 2014]
    disps = [1, 10, 100]

    # System-specific observation model parameters.
    gpss_bgs = [7, 9, 11, 13]
    gpss_prs = np.linspace(0.25e-4, 10e-4, num=40)
    mmds_bgs = [0.0043, 0.0053, 0.0063, 0.0073]
    mmds_prs = np.linspace(0.2e-7, 8e-7, num=40)
    vdh_bgs = {2009: 14, 2010: 15, 2011: 32, 2012: 28, 2013: 32, 2014: 46,
               2015: 58}
    vdh_prs = np.linspace(0.25e-3, 10e-3, num=40)

    # Determine the systems from which to generate forecasts.
    fx_gpss = args['gpss']
    fx_mmds = args['mmds']
    fx_vdh = args['vdh']

    args['gpss'], args['mmds'], args['vdh'] = False, False, False

    if fx_gpss:
        # Generate forecasts from the GPSS data.
        args['gpss'] = True
        for year in years:
            args['year'] = year
            for bg in gpss_bgs:
                args['gpss_bg'] = bg
                for pr in gpss_prs:
                    args['gpss_pr'] = pr
                    for disp in disps:
                        args['gpss_disp'] = disp
                        if counter == choice:
                            run(args)
                        counter = (counter + 1) % partitions
        args['gpss'] = False

    if fx_mmds:
        # Generate forecasts from the MMDS data.
        args['mmds'] = True
        for year in years:
            args['year'] = year
            for bg in mmds_bgs:
                args['mmds_bg'] = bg
                for pr in mmds_prs:
                    args['mmds_pr'] = pr
                    for disp in disps:
                        args['mmds_disp'] = disp
                        if counter == choice:
                            run(args)
                        counter = (counter + 1) % partitions
        args['mmds'] = False

    if fx_vdh:
        # Generate forecasts from the VDH data.
        args['vdh'] = True
        for year in years:
            args['year'] = year
            year_bg = vdh_bgs[year]
            for bg in [year_bg, year_bg - 5, year_bg + 5]:
                args['vdh_bg'] = bg
                for pr in vdh_prs:
                    args['vdh_pr'] = pr
                    for disp in disps:
                        args['vdh_disp'] = disp
                        if counter == choice:
                            run(args)
                        counter = (counter + 1) % partitions
        args['vdh'] = False


if __name__ == '__main__':
    sys.exit(main())
